import logging

import idaapi
import idc

logger = logging.getLogger(__name__)


def analyze() -> None:
    """ Run the IDA analysis and load Objective-C plugins. """
    logger.info('Analyzing...')
    idc.auto_mark_range(0, idc.BADADDR, idc.AU_FINAL)
    idc.auto_wait()

    for action in [1, 4, 5]:
        # 1. Analyze objc segments
        # 4. Analyze NSConcreteGlobalBlock objects
        # 5. Analyze NSConcreteStackBlock objects
        idc.load_and_run_plugin('objc', action)

    idc.auto_wait()
    logger.info('Analysis completed.')


def dscu_load_module(module: str) -> None:
    """ Load a single module into IDA. """
    node = idaapi.netnode()
    node.create('$ dscu')
    node.supset(2, module)
    idc.load_and_run_plugin('dscu', 1)
    idc.auto_wait()


def dscu_load_modules(modules: list[str]) -> None:
    """ Load multiple modules into IDA. """
    for module in modules:
        logger.info(f'Loading module: {module}')
        dscu_load_module(module)
    analyze()


def dscu_load_region(ea: int) -> None:
    """ Load a single memory region into IDA. """
    node = idaapi.netnode()
    node.create('$ dscu')
    node.altset(3, ea)
    idc.load_and_run_plugin('dscu', 2)
    idc.auto_wait()


def dscu_load_regions(regions: list[int]) -> None:
    """ Load multiple memory regions into IDA. """
    for region in regions:
        logger.debug(f'Loading region: {hex(region)}')
        dscu_load_region(region)
    analyze()



import logging
import re
from dataclasses import dataclass
from pathlib import Path

from plumbum import local

from anubis.collectors.base_collector import BaseCollector
from anubis.common import EnhancedDict

logger = logging.getLogger(__name__)

STRINGS = local['strings']


@dataclass
class StringsCollector(BaseCollector):
    """ A collector that processes files using the `strings` command to extract patterns. """

    files: list[dict[str, str]]

    @classmethod
    def from_rule(cls, rule: list[dict[str, str]]):
        return cls(rule)

    async def collect(self, root_path: Path, output_dir: Path):
        res = EnhancedDict()
        for file in self.files:
            path, pattern = file['path'], file['pattern']
            logger.info(f'Collecting {path}...')
            data = STRINGS(str(root_path / path))
            res[path] = list(set(re.findall(pattern, data)))
        res.save_to_yaml(output_dir / 'strings.yaml')



import dataclasses
import logging
import plistlib
import tempfile
from pathlib import Path

import yaml
from plumbum import local

logger = logging.getLogger(__name__)

YQ = local['/opt/homebrew/bin/yq']


@dataclasses.dataclass()
class PlistCollector:
    """ Collector for processing plist files and converting them to YAML format. """

    rules: list[dict[str, str]]

    @classmethod
    def from_rule(cls, rule: list[dict[str, str]]) -> 'PlistCollector':
        return cls(rule)

    async def collect(self, root_path: Path, output_dir: Path) -> None:
        output_dir = output_dir / 'plist'
        output_dir.mkdir(exist_ok=True)

        for rule in self.rules:
            file = rule['file']
            file_path = root_path / file
            if not file_path.exists():
                logger.warning(f'File {file_path} does not exist')
                continue
            logger.info(f'Processing plist file: {file}')

            with file_path.open('rb') as plist_file:
                data = plistlib.load(plist_file)

            yq_query = rule.get('yq_query')
            if yq_query:
                logger.info(f'Running yq query: {yq_query}')
                with tempfile.NamedTemporaryFile(mode='w+', delete=False) as fp:
                    yaml.dump(data, fp, sort_keys=True, indent=4)
                    data = YQ(yq_query, fp.name).splitlines()

            output_path = output_dir / f'{file_path.name}.yaml'
            with output_path.open('w') as output_file:
                yaml.dump(data, output_file, sort_keys=True, indent=4)

            logger.info(f'Converted plist to YAML: {output_path}')



import logging
import re
from dataclasses import dataclass
from enum import Enum, auto
from typing import Callable, Optional

import ida_xref
import idautils
import idc

from anubis.ida_rpyc.regs_tracker.disassembler import DisplacementOperand, ImmediateOperand, Instruction, Register, \
    RegOperand
from anubis.ida_rpyc.regs_tracker.exceptions import UnmappedMemoryError

logger = logging.getLogger(__name__)

OBJC_RETAIN_PATTERN = re.compile(r'_objc_retain_(x\d+)')
FUNC_NAME_EXCLUDE = re.compile(r'\.cold\.\d+$')
SEGMENT_BLACKLIST = ['__auth_stubs']

REGISTER_WIDTH = 8
MAX_DEPTH = 10


class TrackerStatus(Enum):
    TRACKING = auto(),
    VALUE_FOUND = auto(),
    MAX_DEPTH = auto(),
    NO_XREFS = auto(),
    DATA_XREFS = auto(),
    NOT_SUPPORTED = auto(),


def handle_objc_retain(func_name: str) -> Register:
    """ Extracts source register from the given Objective-C retain function name. """
    reg = OBJC_RETAIN_PATTERN.match(func_name)
    if reg:
        return Register.from_str(reg.group(1))
    raise ValueError(f'Invalid function name: {func_name}')


BL_HANDLERS_REGEX: dict[re.Pattern, Callable[[str], Register]] = {
    re.compile(r'^_objc_msgSend\$stringWithFormat:$'): lambda func_name: Register.from_str('x2'),
    re.compile(r'^_objc_retain_'): handle_objc_retain,
}


def get_bl_handler(op_name: str) -> Optional[Callable[[str], Register]]:
    """ Retrieves the appropriate handler for a given branch instruction name. """
    for pattern, handler in BL_HANDLERS_REGEX.items():
        if pattern.match(op_name):
            return handler
    return None


@dataclass
class LogEntry:
    instruction: Instruction
    register: Register
    status: TrackerStatus
    comment: str = ''

    def __str__(self) -> str:
        operands_str = ', '.join(map(str, self.instruction.operands))
        address = f'{hex(self.instruction.ea):<12}'
        mnemonic = f'{self.instruction.mnemonic:<10}'
        operands = f'{operands_str:<60}'
        register = f'{str(self.register):<3}'
        # Use the enum's name for a concise status representation.
        status = f'{self.status.name:<15}'
        comment = self.comment or ''
        return f'{address}: {mnemonic} {operands} | {register} | {status} | {comment}'


class RegisterTracker:
    def __init__(self, target_register: Register, instruction: Instruction) -> None:
        self.target_register = target_register
        self.current_instruction = instruction
        self._current_register = self.target_register
        self._status = TrackerStatus.TRACKING
        self._value = None
        self.trace_log: list[LogEntry] = []

    def __repr__(self) -> str:
        """Custom representation of the register tracker."""
        return (f'RegisterTracker(target_register={self.target_register}, current_register={self.current_register},'
                f'status={self.status}, value={self.value})')

    def clone(self) -> 'RegisterTracker':
        """Creates a copy of the current register tracker."""
        cloned = RegisterTracker(self.target_register, self.current_instruction)
        cloned._current_register = self._current_register
        cloned._value = self._value
        cloned._status = self.status
        cloned.trace_log = self.trace_log[:]
        return cloned

    def record(self, comment: str) -> None:
        """Records an event in the tracker log."""
        self.trace_log.append(LogEntry(self.current_instruction, self._current_register, self.status, comment))

    @property
    def current_register(self) -> Register:
        return self._current_register

    @current_register.setter
    def current_register(self, value: Register) -> None:
        self.record(f'Changing register {self.current_register} -> {value}')
        self._current_register = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value) -> None:
        self._value = value

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value) -> None:
        self._status = value
        self.record(f'Status changed: {value}')

    @property
    def bt(self) -> str:
        """Provides a backtrace of the register tracking."""
        trace_lines = []
        for entry in reversed(self.trace_log):
            trace_lines.append(f'    {str(entry)}')
        return '\n'.join(trace_lines)

    def handle_current_instruction(self):
        instruction = self.current_instruction
        current_register = self.current_register
        # Handle 'BL' instructions, which often correspond to function calls.
        # Check if the mnemonic is 'BL' and if the current register is x0, which is commonly used for function arguments.
        if instruction.mnemonic == 'BL' and current_register == Register.from_str('x0') \
                and self.target_register != current_register:
            op1 = instruction.operands[0]

            # Ensure the operand is mapped to valid memory.
            if not op1.is_mapped:
                raise UnmappedMemoryError(op1.addr)

            # Attempt to find a handler for the called function.
            handler = get_bl_handler(op1.name)
            if handler is None:
                self.record('Unknown after call')
            else:
                self.current_register = handler(op1.name)

        # Check if the instruction writes to the register we are tracking.
        elif len(instruction.operands) > 1 and instruction.operands[0].reg == current_register:
            op2 = instruction.operands[1]

            # Handle MOV instructions.
            if instruction.mnemonic == 'MOV':
                if isinstance(op2, ImmediateOperand):
                    self.status = TrackerStatus.VALUE_FOUND
                    self.current_instruction = instruction
                # If the source is another register, update register to track
                elif isinstance(op2, RegOperand):
                    self.current_register = op2.reg

            # Handle ADRP/ADRL instructions.
            elif instruction.mnemonic in ['ADR', 'ADRP', 'ADRL']:
                # If the source is an immediate value,resolve and return.
                if isinstance(op2, ImmediateOperand):
                    self.status = TrackerStatus.VALUE_FOUND
                    self.current_instruction = instruction

            # Handle LDR instructions.
            elif instruction.mnemonic == 'LDR' and isinstance(op2, DisplacementOperand):
                # If the operand is a displacement (e.g., [base + offset]), track the base register.
                if op2.offset != 0:
                    self.status = TrackerStatus.NOT_SUPPORTED
                    self.value = hex(instruction.ea)
                else:
                    self.current_register = op2.base
            else:
                self.status = TrackerStatus.NOT_SUPPORTED
                self.value = hex(instruction.ea)

            if self.status == TrackerStatus.VALUE_FOUND:
                self.value = op2.resolve_data_value()

    @staticmethod
    def track_register_at(register: Register, start_address: int, max_depth: int = MAX_DEPTH) -> list['RegisterTracker']:
        """ Initiates tracking of a register starting from a given address. """
        instruction = Instruction.from_ea(start_address)
        tracker = RegisterTracker(register, instruction)
        tracker.record(f'Start Tracking {register} at {start_address}')
        return tracker._track_register_at(max_depth)

    def _track_register_at(self, max_depth: int = MAX_DEPTH) -> list['RegisterTracker']:
        """ Recursively tracks the register. """
        tracker = self
        result: list[RegisterTracker] = []

        # Base case: if the recursion depth has reached zero, stop tracking.
        current_instruction_ea = tracker.current_instruction.ea
        if max_depth == 0:
            tracker.status = TrackerStatus.MAX_DEPTH
            tracker.value = hex(current_instruction_ea)
            return [tracker]

        # Handle current instruction
        tracker.handle_current_instruction()
        if tracker.status != TrackerStatus.TRACKING:
            return [tracker]

        # At this point, the instruction didn't resolve the register value directly.
        # So, follow the xrefs (cross-references) to see where the value might come from.

        # Get all xrefs pointing to the current instruction's address.
        xrefs = list(idautils.XrefsTo(current_instruction_ea))
        if not xrefs:
            # If no cross-references are found, mark the value as unresolved.
            tracker.status = TrackerStatus.NO_XREFS
            tracker.value = hex(current_instruction_ea)
            return [tracker]

        # If there are cross-references, recursively trace through each one.
        for xref in xrefs:
            # Clone the trace entry for each xref to track distinct execution paths.
            new_tracker = tracker.clone()

            if not xref.iscode:
                # If the xref is not from code, it's likely a data reference.
                new_tracker.status = TrackerStatus.DATA_XREFS
                new_tracker.value = hex(xref.frm)
                result.append(new_tracker)
                continue

            new_tracker.current_instruction = Instruction.from_ea(xref.frm)

            if xref.type != ida_xref.fl_F:
                # If it's not a standard flow xref (BL), log the tracking action and decrease depth.
                new_tracker.record(f'Tracking x-refs to {hex(current_instruction_ea)}')
                result += new_tracker._track_register_at(max_depth - 1)
            else:
                # If it is a flow xref, continue tracking with the same depth.
                result += new_tracker._track_register_at(max_depth)

        return result


def get_function_addresses_by_name(name: str) -> list[tuple[int, str]]:
    """ Retrieves addresses of functions that match the given name. """
    matches = []
    for ea in idautils.Functions():
        func_name = idc.get_func_name(ea)
        if name in func_name:
            seg = idc.get_segm_name(ea)
            logger.info(seg)
            parts = seg.split(':')
            if len(parts) > 1 and parts[1] in SEGMENT_BLACKLIST:
                logger.info(f'Function {func_name} is blacklisted: {seg}')
                continue
            if FUNC_NAME_EXCLUDE.search(func_name):
                logger.info(f'Function {func_name} is blacklisted')
                continue
            matches.append((ea, func_name))

    return matches


def track_function_registers(name: str, registers: list[str], max_depth: int = MAX_DEPTH) -> dict[str, list[RegisterTracker]]:
    """ Tracks the specified registers within functions that match the given name. """
    func = get_function_addresses_by_name(name)
    results: dict[str, list[RegisterTracker]] = {}
    for ea, func_name in func:
        print(f'Function: {func_name} @ 0x{ea:X}')
        for r in registers:
            results[r] = RegisterTracker.track_register_at(Register.from_str(r), ea, max_depth)
    return results



import logging
import re
from dataclasses import dataclass
from pathlib import Path

from anubis.collectors.base_collector import BaseCollector
from anubis.common import DSC_PATH, IPSW, EnhancedDict, get_dylibs_names_from_dsc

logger = logging.getLogger(__name__)


@dataclass
class ClassDumpCollector(BaseCollector):
    """ Collector for extracting class dump information. """

    rules: list[dict[str, str]]

    @classmethod
    def from_rule(cls, rule: list[dict[str, str]]) -> 'ClassDumpCollector':
        return cls(rule)

    async def collect(self, root_path: Path, output_dir: Path) -> None:
        res = EnhancedDict()
        dylibs = get_dylibs_names_from_dsc(root_path)
        for rule in self.rules:
            path, pattern = rule['path'], rule['pattern']
            logger.info(f'Collecting {path}...')
            if f'{path}' in dylibs:
                data = IPSW('class-dump', str(root_path / DSC_PATH), path)
            else:
                data = IPSW('class-dump', str(root_path / path))
            res[path] = re.findall(pattern, data)

        res.save_to_yaml(output_dir / 'class-dump.yaml')



import importlib
import logging

import dsc_stuff
import ida_auto
import ida_loader
import ida_pro
import idc
import rpyc

from anubis.common import MAX_DEPTH
from anubis.ida_rpyc.regs_tracker.ida_reg_tracker import track_function_registers

logger = logging.getLogger(__name__)


class IDATrackService(rpyc.Service):
    """Service for tracking function registers using IDA Pro."""

    def __init__(self) -> None:
        """Initialize the service and wait for IDA to be ready."""
        super().__init__()
        ida_auto.auto_wait()

    def on_connect(self, conn):
        ida_loader.set_database_flag(ida_loader.DBFL_KILL | ida_loader.DBFL_TEMP)

    def on_disconnect(self, conn) -> None:
        """Handle client disconnection by exiting IDA."""
        ida_pro.qexit(0)

    def exposed_import_module(self, mod: str):
        """Dynamically import a module by name."""
        return importlib.import_module(mod)

    def exposed_track_function_registers(self, name: str, registers: list[str], max_depth: int = MAX_DEPTH):
        """Track specified registers in a given function with a maximum search depth."""
        return track_function_registers(name, registers, max_depth)

    def exposed_load_modules(self, modules: list[str]) -> None:
        """Load specified modules using dsc_stuff."""
        dsc_stuff.dscu_load_modules(modules)

    def exposed_load_regions(self, regions: list[int]) -> None:
        """Load specified memory regions using dsc_stuff."""
        dsc_stuff.dscu_load_regions(regions)


if __name__ == '__main__':
    server = rpyc.OneShotServer(
        service=IDATrackService,
        port=idc.ARGV[1],
        protocol_config={
            'allow_public_attrs': True,
            'import_custom_exceptions': True,
        }
    )
    rpyc.lib.setup_logger()
    server.start()



import logging
from dataclasses import dataclass
from pathlib import Path

from anubis.collectors.base_collector import BaseCollector
from anubis.common import EnhancedDict, run_ripgrep

logger = logging.getLogger(__name__)


@dataclass
class RgCollector(BaseCollector):
    """ Collector that searches for specific patterns within a given directory using ripgrep. """

    patterns: list[str]

    @classmethod
    def from_rule(cls, rule: dict) -> 'RgCollector':
        return cls(rule['patterns'])

    async def collect(self, root_path: Path, output_dir: Path):
        res = await run_ripgrep(root_path, self.patterns)
        EnhancedDict(res).save_to_yaml(output_dir / 'rg.yaml')



import logging
import re
from dataclasses import dataclass
from pathlib import Path

from plumbum import local

from anubis.collectors.base_collector import BaseCollector
from anubis.common import EnhancedDict

logger = logging.getLogger(__name__)
NM = local['nm']


@dataclass
class NmCollector(BaseCollector):
    """ A collector that extracts symbol information using nm. """

    modules: list[dict[str, str]]

    @classmethod
    def from_rule(cls, rule: list[dict[str, str]]) -> 'NmCollector':
        return cls(rule)

    async def collect(self, root_path: Path, output_dir: Path) -> None:
        res = EnhancedDict()
        for module in self.modules:
            path, pattern = module['path'], module['pattern']
            logger.info(f'Collecting {path} with pattern {pattern}')
            data = NM(root_path / path)
            res[path] = re.findall(pattern, data)
        res.save_to_yaml(output_dir / 'nm.yaml')



import logging
import os
import socket
import subprocess
import tempfile
import time
from pathlib import Path
from typing import Any, Optional

import rpyc

from anubis.common import MAX_DEPTH
from anubis.ida_rpyc.regs_tracker.exceptions import UnmappedMemoryError

SERVER_SCRIPT = Path(__file__).parent / 'ida_server.py'

logger = logging.getLogger(__name__)
rpyc.core.vinegar._generic_exceptions_cache[
    'anubis.ida_rpyc.regs_tracker.exceptions.UnmappedMemoryError'
] = UnmappedMemoryError


class MachoIdb:
    """ Wrapper for interacting with IDA to analyze Mach-O binaries. """

    def __init__(self, ida_path: str, binary_path: str, ftype: str = 'Mach-O file', env: Optional[dict] = None) -> None:
        self.ida_path = ida_path
        self.binary_path = binary_path
        self.ftype = ftype
        self.env = {} if env is None else env
        self._conn = None

    def start(self) -> None:
        """ Launch IDA and establish a rpyc connection. """
        with socket.socket() as s:
            s.bind(('', 0))
            port = s.getsockname()[1]

        tempidb = tempfile.NamedTemporaryFile()
        command = f'"{self.ida_path}" -A -T"{self.ftype}" -o"{tempidb.name}" -S"{SERVER_SCRIPT} {port}" "{self.binary_path}"'
        self.env.update(os.environ.copy())

        p = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=self.env)

        while self._conn is None:
            if p.poll() is not None:
                raise Exception(
                    f'IDA failed to start: return code {p.poll()}\n'
                    f'Command: {command}\n'
                    f'=============== STDOUT ===============\n{p.stdout.read().decode()}'
                    f'=============== STDERR ===============\n{p.stderr.read().decode()}'
                )
            try:
                self._conn = rpyc.connect('localhost', port, config={'sync_request_timeout': 60 * 60 * 24})
            except ConnectionError:
                time.sleep(1)

    def stop(self) -> None:
        """ Close the rpyc connection. """
        self._conn.close()

    def track_function_registers(self, name: str, registers: tuple, max_depth: int = MAX_DEPTH) -> dict:
        """ Track function register values and return them as a Python object. """
        raw_results = self._conn.root.track_function_registers(name, registers, max_depth)
        results = {register: [(x.status.name, x.value) for x in value] for register, value in raw_results.items()}
        return results

    def __enter__(self) -> 'MachoIdb':
        """ Start the connection when entering a context. """
        self.start()
        return self

    def __exit__(self, exc_type: Optional[type], exc_value: Optional[Exception], traceback: Optional[Any]) -> None:
        """ Stop the connection when exiting a context. """
        self.stop()


class DscIdb(MachoIdb):
    """ Wrapper for analyzing Apple DYLD cache modules in IDA. """

    def __init__(self, ida_path: str, module: str, dsc_path: str) -> None:
        super().__init__(ida_path, dsc_path, 'Apple DYLD cache for arm64e (select module(s))',
                         {'IDA_DYLD_CACHE_MODULE': module})

    def track_function_registers(self, name: str, registers: tuple, max_depth: int = MAX_DEPTH) -> dict:
        """ Track function registers, handling unmapped memory errors by loading required regions. """
        ret = None
        loaded_regions = set()

        while ret is None:
            try:
                ret = super().track_function_registers(name, registers, max_depth)
            except UnmappedMemoryError as e:
                unmapped_address = e.address
                logger.warning(f'Got UnmappedMemoryError at: {hex(unmapped_address)}, trying to load region')

                if unmapped_address in loaded_regions:
                    raise e

                self._conn.root.load_regions([unmapped_address])
                loaded_regions.add(unmapped_address)
                logger.info(f'Region loaded {hex(unmapped_address)}, retrying...')

        return ret



import asyncio
import logging
import traceback
from concurrent.futures import ProcessPoolExecutor
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Union

from tqdm import tqdm

from anubis.collectors.base_collector import BaseCollector
from anubis.common import DSC_PATH, EnhancedDict, get_dylibs_names_from_dsc, run_ripgrep
from anubis.ida_rpyc.ida_client import DscIdb, MachoIdb

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

IDA_PATH = '/Applications/IDA Professional 9.2.app/Contents/MacOS/ida'


@dataclass(frozen=True)
class SearchRule:
    """ A search rule defining a function and associated registers. """
    function: str
    registers: tuple[str]


@dataclass
class RegsTrackerTaskMacho:
    """ A task for tracking registers in Mach-O files using IDA. """
    root_path: str
    file_path: str
    search_rules: set[SearchRule]

    @contextmanager
    def idb_context(self, ida_path: str):
        """ Context manager for regular tasks using MachoIdb. """
        with MachoIdb(ida_path, f'{self.root_path}/{self.file_path}') as idb:
            yield idb


@dataclass
class RegsTrackerTaskDsc(RegsTrackerTaskMacho):
    """ A task for tracking registers in a DSC file using IDA. """
    dsc_path: str

    @contextmanager
    def idb_context(self, ida_path: str):
        """ Context manager for DSC tasks using DscIdb. """
        with DscIdb(ida_path, self.file_path, self.dsc_path) as idb:
            yield idb


def reg_tracker_worker(task: Union[RegsTrackerTaskDsc, RegsTrackerTaskDsc]):
    """ Worker function for processing a register tracking task. """
    try:
        results = {}
        with task.idb_context(IDA_PATH) as idb:
            for rule in task.search_rules:
                results[rule.function] = idb.track_function_registers(rule.function, rule.registers)
        return task.file_path, results
    except Exception as e:
        logger.error(f'Error for task {task} traceback: {e}\n{traceback.format_exc()}')
        raise


@dataclass
class RegsTrackerCollector(BaseCollector):
    """ Collector for tracking register values of a given function. """
    ida_path: str
    search_rules: dict[str, tuple[str]]

    @classmethod
    def from_rule(cls, rule) -> 'RegsTrackerCollector':
        """ Create an IDBCollector instance from a rule definition. """
        return cls(IDA_PATH, {p['func_pattern']: tuple(p['registers']) for p in rule})

    async def collect(self, root_path: Path, output_dir: Path) -> None:
        matches = await self._find_matches(root_path)
        tasks = await self._create_tasks(root_path, matches.invert())
        results = await self._process_tasks(tasks)

        inverted_results = EnhancedDict(results).invert()
        output_dir = output_dir / 'regs_tracker'
        output_dir.mkdir(parents=True, exist_ok=True)
        for func, results in inverted_results.items():
            output_file = output_dir / f'{func.replace(":", "_")}.yaml'
            EnhancedDict(results).save_to_yaml(output_file)

    async def _create_tasks(self, root_path: Path, matches: EnhancedDict):
        """ Create tasks for register tracking based on matched patterns. """
        tasks = []
        dylibs = get_dylibs_names_from_dsc(root_path)
        for k, v in matches.items():
            if f'/{k}' in dylibs:
                task = RegsTrackerTaskDsc(str(root_path), f'/{k}', set([(SearchRule(p, self.search_rules[p])) for p in v]),
                                          dsc_path=str(root_path / DSC_PATH))
            else:
                task = RegsTrackerTaskMacho(str(root_path), k, set([(SearchRule(p, self.search_rules[p])) for p in v]))
            tasks.append(task)

        return tasks

    @staticmethod
    async def _process_tasks(tasks: list[Union[RegsTrackerTaskMacho, RegsTrackerTaskDsc]]):
        """ Process tracking tasks using a process pool. """
        loop = asyncio.get_running_loop()
        results = {}
        with ProcessPoolExecutor() as executor:
            futures = [loop.run_in_executor(executor, reg_tracker_worker, task) for task in tasks]
            for future in tqdm(asyncio.as_completed(futures), total=len(tasks), desc='Processing regs-tracker tasks'):
                file_path, result = await future
                results[file_path] = result
        return results

    async def _find_matches(self, root_path: Path) -> EnhancedDict:
        """ Find matches by running ripgrep with the given search rules. """
        return await run_ripgrep(root_path, list(self.search_rules.keys()))



import asyncio
import logging
from pathlib import Path

from idascript import IDA

from anubis.collectors.base_collector import FilesCollector

logger = logging.getLogger(__name__)


class BinExportCollector(FilesCollector):
    """ Collector for exporting binary analysis results using IDA. """

    async def collect(self, root_path: Path, output_dir: Path) -> None:
        output_dir = output_dir / 'binexport'
        output_dir.mkdir(exist_ok=True)

        for file in self.files:
            logger.info(f'Collecting {file}...')
            output_file = output_dir / Path(file.name).with_suffix('.BinExport')

            ida = IDA(
                root_path / file,
                script_file=None,
                script_params=[
                    'BinExportAutoAction:BinExportBinary',
                    f'BinExportModule:{output_file}',
                ],
            )
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(None, ida.start)
            await loop.run_in_executor(None, ida.wait)

            if output_file.exists():
                logger.info(f'Done {output_file}...')



import logging
from dataclasses import dataclass, field
from typing import List, Optional, Union

import ida_bytes
import ida_idp
import ida_ua
import idc

logger = logging.getLogger(__name__)

REGISTER_WIDTH = 8


def extract_objc_string(nsstring_addr: int) -> Optional[str]:
    """Extracts an Objective-C string from memory if present."""
    c_string_addr = ida_bytes.get_64bit(nsstring_addr + 0x10)
    if c_string_addr:
        string = idc.get_strlit_contents(c_string_addr)
        if string:
            return string.decode('utf-8')
    return None


def extract_c_cstring(c_string_address: int) -> Optional[str]:
    """Extracts a C string from memory if present."""
    string = idc.get_strlit_contents(c_string_address)
    if string:
        return string.decode('utf-8')
    return None


@dataclass
class Register:
    """Represents a CPU register."""
    num: int

    @staticmethod
    def from_str(name: str) -> 'Register':
        """Creates a Register instance from a string name."""
        return Register(ida_idp.str2reg(name))

    def __str__(self) -> str:
        """Returns the register name as a string."""
        return ida_idp.get_reg_name(self.num, REGISTER_WIDTH)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Register) and self.num == other.num

    def __gt__(self, other: 'Register') -> bool:
        return self.num > other.num

    def __lt__(self, other: 'Register') -> bool:
        return self.num < other.num

    def __repr__(self) -> str:
        return f'Register(reg={str(self)})'


@dataclass
class Operand:
    """Base class for instruction operands."""
    pass


@dataclass
class RegOperand(Operand):
    """Represents a register operand."""
    reg: Register

    def __repr__(self) -> str:
        return f'RegOperand(reg={self.reg})'


@dataclass
class UnknownOperand(Operand):
    """Represents an unknown operand type."""
    op_type: int

    def __repr__(self) -> str:
        return f'UnknownOperand(type={self.op_type})'


@dataclass
class MemoryOperand(Operand):
    """Represents a memory address operand."""
    addr: int

    @property
    def name(self) -> str:
        """Returns the name of the memory location."""
        return idc.get_name(self.addr)

    @property
    def is_mapped(self) -> bool:
        """Checks if the memory address is mapped."""
        return idc.is_mapped(self.addr)

    def __repr__(self) -> str:
        return f'MemoryOperand(addr=0x{self.addr:X},is_mapped={self.is_mapped})'


@dataclass
class ImmediateOperand(Operand):
    """Represents an immediate operand (constant value)."""
    value: int

    def __repr__(self) -> str:
        return f'ImmediateOperand(value=0x{self.value:X})'

    def resolve_data_value(self) -> Union[str, int]:
        """Resolves the immediate value to a string or name if possible."""
        return extract_objc_string(self.value) or extract_c_cstring(self.value) or idc.get_name(self.value) or self.value


@dataclass
class DisplacementOperand(Operand):
    """Represents a displacement operand with a base register and an offset."""
    base: Register
    offset: int

    def __repr__(self) -> str:
        return f'DisplacementOperand(base={self.base}, offset=0x{self.offset:X})'


@dataclass
class IndexedDisplacementOperand(Operand):
    """Represents an operand with a base register, index register, and scale factor."""
    base: Register
    index: Register
    scale: int

    def __repr__(self) -> str:
        return f'IndexedDisplacementOperand(base={self.base}, index={self.index}, scale={self.scale})'


Operands = Union[
    IndexedDisplacementOperand, DisplacementOperand, ImmediateOperand, MemoryOperand, RegOperand, UnknownOperand]


@dataclass
class Instruction:
    """Represents a disassembled instruction."""
    ea: int
    size: int
    itype: int
    mnemonic: str
    operands: List[Operands] = field(default_factory=list)

    @staticmethod
    def from_ea(ea: int) -> Optional['Instruction']:
        """Decodes an instruction from the given address."""
        insn = ida_ua.insn_t()
        decoded_len = ida_ua.decode_insn(insn, ea)
        if decoded_len == 0:
            return None

        operands = [
            create_operand(op)
            for i, op in enumerate(insn.ops)
            if op.type != ida_ua.o_void
        ]

        return Instruction(
            ea=ea,
            size=insn.size,
            itype=insn.itype,
            mnemonic=idc.print_insn_mnem(ea),
            operands=operands,
        )

    def __repr__(self) -> str:
        ops_str = ', '.join(repr(op) for op in self.operands)
        return f'<Instruction 0x{self.ea:X} mnemonic="{self.mnemonic}" [{ops_str}]>'


def create_operand(op: ida_ua.op_t) -> Operands:
    """Creates an operand instance based on the operand type."""
    if op.type == ida_ua.o_reg:
        return RegOperand(reg=Register(op.reg))
    elif op.type in {ida_ua.o_mem, ida_ua.o_far, ida_ua.o_near}:
        return MemoryOperand(addr=op.addr)
    elif op.type == ida_ua.o_imm:
        return ImmediateOperand(value=op.value)
    elif op.type == ida_ua.o_displ:
        base_reg = Register(op.phrase)
        if op.specflag1:
            index_reg = Register(op.specflag1)
            scale = 1 << op.specflag2
            return IndexedDisplacementOperand(base=base_reg, index=index_reg, scale=scale)
        else:
            return DisplacementOperand(base=base_reg, offset=op.addr)
    elif op.type == ida_ua.o_phrase:
        base_reg = Register(op.phrase)
        index_reg = Register(op.specflag1)
        scale = 1 << op.specflag2 if op.specflag2 else 1
        return IndexedDisplacementOperand(base=base_reg, index=index_reg, scale=scale)
    else:
        return UnknownOperand(op_type=op.type)



import asyncio
import json
import logging
import re
from collections import UserDict
from pathlib import Path
from typing import Any

import yaml
from plumbum import local

IPSW = local['/opt/homebrew/bin/ipsw']
DSC_PATH = Path('System/Cryptexes/OS/System/Library/Caches/com.apple.dyld/dyld_shared_cache_arm64e')
EXTENSIONS_BLACKLIST = ('.i64', '.dyldlinkedit', '.symbols', '.BinExport', '.id0')
DYLD_SHARED_CACHE_PATTERN = re.compile(r'^dyld_shared_cache_arm64e(\..+)?$')
MAX_DEPTH = 5
RG_PATH = '/opt/homebrew/bin/rg'
YAML_UNDEFINED_TAG = 'tag:yaml.org,2002:str'

logger = logging.getLogger(__name__)


class AnubisYamlDumper(yaml.Dumper):
    """ Anubis YAML dumper that registers specific representers. """
    pass


class EnhancedDict(UserDict):
    """ A dictionary subclass that provides additional functionality such as inversion and YAML saving. """

    def invert(self) -> 'EnhancedDict':
        """ Invert the dictionary, swapping keys and values while maintaining structure. """
        inverted = {}
        for outer_key, value in self.data.items():
            if isinstance(value, dict):
                for inner_key, inner_value in value.items():
                    if inner_key not in inverted:
                        inverted[inner_key] = {outer_key: inner_value}
                    else:
                        inverted[inner_key][outer_key] = inner_value
            elif isinstance(value, list):
                for item in value:
                    inverted.setdefault(item, []).append(outer_key)
            else:
                inverted.setdefault(value, []).append(outer_key)
        return EnhancedDict(inverted)

    def save_to_yaml(self, output_file: Path) -> None:
        """ Save the dictionary to a YAML file. """
        with output_file.open('w', encoding='utf-8') as file:
            data_str_keys = {str(k): v for k, v in self.data.items()}
            yaml.dump(data_str_keys, file, Dumper=AnubisYamlDumper, sort_keys=True, indent=4)

    def __str__(self) -> str:
        """ String representation of the dictionary. """
        return '\n'.join(f'{k}: {v}' for k, v in self.data.items())


def anubis_dumper_representeter(dumper: yaml.Dumper, data: Any) -> yaml.Node:
    """ Anubis YAML representer for EnhancedDict and related data structures. """
    if isinstance(data, dict):
        return dumper.represent_dict(sorted(data.items()))
    elif isinstance(data, list):
        return dumper.represent_list(sorted(data, key=str))
    elif isinstance(data, tuple):
        return dumper.represent_scalar(YAML_UNDEFINED_TAG, ', '.join(map(str, data)))
    return dumper.represent_data(data)


AnubisYamlDumper.add_representer(EnhancedDict, anubis_dumper_representeter)
AnubisYamlDumper.add_representer(dict, anubis_dumper_representeter)
AnubisYamlDumper.add_representer(list, anubis_dumper_representeter)
AnubisYamlDumper.add_representer(tuple, anubis_dumper_representeter)


def get_dylibs_names_from_dsc(root_path: Path) -> list[str]:
    """ Retrieve a list of dylib names from the dyld_shared_cache. """
    dsc_related = IPSW('dsc', 'info', '-l', '--json', str(root_path / DSC_PATH))
    return [d['name'] for d in json.loads(dsc_related)['dylibs']]


async def run_ripgrep(root_path: Path, patterns: list[str],
                      excludes_extensions: list[str] = EXTENSIONS_BLACKLIST) -> EnhancedDict:
    """ Run ripgrep to find patterns in files within a directory. """
    cmd = [
        RG_PATH,
        '-a',  # Search binary files as if they were text
        '--no-heading',  # Remove filepath header for each result set
        '-o',  # Print only the matched (non-empty) parts of the matching line
        '-e',
        '|'.join(patterns),
        str(root_path)
    ]

    patterns_fmt = '\n\t'.join(patterns)
    logger.info(f'Running ripgrep on patterns:\n\t{patterns_fmt}')

    process = await asyncio.create_subprocess_exec(
        *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )
    stdout, stderr = await process.communicate()

    if stderr:
        logger.error(f'Error: {stderr.decode("utf-8")}')

    res = EnhancedDict()
    for line in set(stdout.decode('utf-8').splitlines()):
        matched_file, pattern = line.split(':', 1)
        file_name = Path(matched_file).relative_to(root_path)

        if excludes_extensions and file_name.suffix in excludes_extensions:
            continue
        if DYLD_SHARED_CACHE_PATTERN.match(file_name.name):
            continue

        file_name = str(file_name)
        res.setdefault(pattern, []).append(file_name)

    return res



import logging
import re
from dataclasses import dataclass
from pathlib import Path

from anubis.collectors.base_collector import BaseCollector
from anubis.common import DSC_PATH, IPSW, EnhancedDict, get_dylibs_names_from_dsc

logger = logging.getLogger(__name__)


@dataclass
class ObjcProtocolSelectors(BaseCollector):
    """ Collector for extracting selectors for given objc protocol. """

    rules: list[dict[str, str]]

    @classmethod
    def from_rule(cls, rule: list[dict[str, str]]) -> 'ObjcProtocolSelectors':
        return cls(rule)

    async def collect(self, root_path: Path, output_dir: Path) -> None:
        """
        Use `class-dump` to generate a header file contents from given executable.
        Then, look for all protocol selectors by searching for all the selectors between `<protocol>`
        to `@end`.

        :param root_path: Extracted IPSW root path.
        :param output_dir: Output directory of where write `objc-protocol-selectors.yaml`.
        :return: None
        """
        res = EnhancedDict()
        dylibs = get_dylibs_names_from_dsc(root_path)
        for rule in self.rules:
            path, protocol = rule['path'], rule['protocol']
            logger.info(f'Collecting {path}...')
            if f'{path}' in dylibs:
                data = IPSW('class-dump', str(root_path / DSC_PATH), path)
            else:
                data = IPSW('class-dump', str(root_path / path))

            protocol_chunk = data.split(f'<{protocol}>')
            if len(protocol_chunk) < 2:
                # No protocol matches were found
                return

            # Iterate each protocol header
            for chunk in protocol_chunk[1:]:
                # Get protocol header body
                protocol_body = chunk.split('@end', 1)[0]

                # Match protocol's selectors
                selectors = re.findall(r'-\[.+?]', protocol_body)

                # Append all matched selectors
                if path not in res:
                    res[path] = []
                res[path] += selectors

            # Make sure the result is always sorted the same way for easier diff-ing
            res[path].sort()

        res.save_to_yaml(output_dir / 'objc-protocol-selectors.yaml')



import logging
from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path

logger = logging.getLogger(__name__)


@dataclass
class BaseCollector:
    """ Abstract base class for collectors. """

    @classmethod
    @abstractmethod
    def from_rule(cls, rule: dict) -> 'BaseCollector':
        """ Create a collector instance from a rule dictionary. """
        pass

    @abstractmethod
    def collect(self, root_path: Path, output_dir: Path) -> None:
        """ Collect data based on the given root path and output directory. """
        pass


@dataclass
class FilesCollector(BaseCollector):
    """ A Collector that handles file-based collection. """

    files: list[Path]

    @classmethod
    def from_rule(cls, rule: dict) -> 'FilesCollector':
        """ Create a FilesCollector instance from a rule dictionary. """
        return cls([Path(path) for path in rule['paths']])

    @abstractmethod
    def collect(self, root_path: Path, output_dir: Path) -> None:
        """ Collect files from the specified root path and save them in the output directory. """
        pass



import logging
import posixpath
from dataclasses import dataclass
from pathlib import Path

from plumbum import local

from anubis.collectors.base_collector import BaseCollector

logger = logging.getLogger(__name__)
XCRUN = local['xcrun']


@dataclass
class SectionCollector(BaseCollector):
    """ A collector that extracts a given section. """

    modules: list[dict[str, str]]

    @classmethod
    def from_rule(cls, rule: list[dict[str, str]]) -> 'SectionCollector':
        return cls(rule)

    async def collect(self, root_path: Path, output_dir: Path) -> None:
        for module in self.modules:
            path, segment, section = module['path'], module['segment'], module['section']
            logger.info(f'Collecting {path}:{segment}:{section}')
            XCRUN('segedit', root_path / path, '-extract', segment, section,
                  output_dir / f'{posixpath.basename(path)}_{segment}_{section}.bin')



import asyncio
import logging
import traceback
from pathlib import Path
from typing import Optional

import click
import coloredlogs
import yaml

from anubis.collectors.binexport import BinExportCollector
from anubis.collectors.class_dump import ClassDumpCollector
from anubis.collectors.nm import NmCollector
from anubis.collectors.objc_protocol_selectors import ObjcProtocolSelectors
from anubis.collectors.plist import PlistCollector
from anubis.collectors.regs_tracker import RegsTrackerCollector
from anubis.collectors.rg import RgCollector
from anubis.collectors.section import SectionCollector
from anubis.collectors.strings import StringsCollector

coloredlogs.install(level=logging.DEBUG)
logging.getLogger('urllib3.connectionpool').disabled = True
logging.getLogger('root').disabled = True
logging.getLogger('plumbum.local').disabled = True
logging.getLogger('asyncio').disabled = True

logger = logging.getLogger('anubis')

TASK_TIMEOUT = 20
COLLECTORS = {'rg': RgCollector, 'binexport': BinExportCollector, 'plist': PlistCollector,
              'class-dump': ClassDumpCollector, 'strings': StringsCollector, 'regs-tracker': RegsTrackerCollector,
              'nm': NmCollector, 'section': SectionCollector, 'objc-protocol-selectors': ObjcProtocolSelectors}


async def collect_task(input_path: Path, output_path: Path, rules: dict, whitelist: Optional[set] = None,
                       blacklist: Optional[set] = None):
    tasks = []
    skip_list = []
    for name, rule in rules.items():
        if (blacklist and name in blacklist) or (whitelist and name not in whitelist):
            skip_list.append(name)
            continue
        try:
            collector = COLLECTORS[name].from_rule(rule)
        except KeyError:
            logger.error(f'Unknown collector: {name}')
            continue
        tasks.append(asyncio.create_task(collector.collect(input_path, output_path), name=name))

    if len(skip_list) > 0:
        logger.info(f'Skipping Collectors: {", ".join(sorted(skip_list))}')

    while tasks:
        done, tasks = await asyncio.wait(tasks, timeout=TASK_TIMEOUT, return_when=asyncio.FIRST_COMPLETED)
        for task in done:
            if task.exception():
                logger.error(
                    f'Collector {task.get_name()} failed with exception: {task.exception()}\n{traceback.format_exc()}')
            else:
                logger.info(f'Collector completed successfully: {task.get_name()}')
        running = ', '.join(sorted([task.get_name() for task in tasks if not task.done()]))
        logger.info(f'Still Running: {running}')

    logger.info('All collectors completed.')


@click.group()
def cli():
    pass


@cli.command()
@click.argument('input_path', type=click.Path(exists=True, dir_okay=True, file_okay=False))
@click.argument('output_path', type=click.Path(exists=False, dir_okay=True, file_okay=False))
@click.argument('rules', type=click.Path(exists=True))
@click.option('collectors', '-c', '--collector', multiple=True)
@click.option('blacklist', '-b', '--blacklist', multiple=True)
def collect(input_path: str, output_path: str, rules: str, collectors: tuple[str], blacklist: tuple[str]):
    input_path, output_path = Path(input_path), Path(output_path)
    output_path.mkdir(exist_ok=True, parents=True)
    with open(rules, 'r') as f:
        data = yaml.safe_load(f)
    wlist = set(collectors) if collectors else None
    blist = set(blacklist) if blacklist else None
    asyncio.run(collect_task(input_path, output_path, data, wlist, blist))


if __name__ == '__main__':
    cli()
